/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2004 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "ButtonFont.h"

#ifndef CLEARTYPE_QUALITY
	#define CLEARTYPE_QUALITY       5
#endif

NameValuePair alignValues[] = {
		{ "left", DT_LEFT },
		{ "center", DT_CENTER },
		{ "right", DT_RIGHT },
		{ 0, 0 }
		};

NameValuePair vertAlignValues[] = {
		{ "top", DT_TOP },
		{ "center", DT_VCENTER },
		{ "bottom", DT_BOTTOM },
		{ 0, 0 }
		};

NameValuePair CharSetConstants[] =
		{
		{ "ANSI_CHARSET", ANSI_CHARSET },
		{ "BALTIC_CHARSET", BALTIC_CHARSET },
		{ "CHINESEBIG5_CHARSET", CHINESEBIG5_CHARSET },
		{ "DEFAULT_CHARSET", DEFAULT_CHARSET },
		{ "EASTEUROPE_CHARSET", EASTEUROPE_CHARSET },
		{ "GB2312_CHARSET", GB2312_CHARSET },
		{ "GREEK_CHARSET", GREEK_CHARSET },
		{ "HANGUL_CHARSET", HANGUL_CHARSET },
		{ "MAC_CHARSET", MAC_CHARSET },
		{ "RUSSIAN_CHARSET", RUSSIAN_CHARSET },
		{ "SHIFTJIS_CHARSET", SHIFTJIS_CHARSET },
		{ "SYMBOL_CHARSET", SYMBOL_CHARSET },
		{ "TURKISH_CHARSET", TURKISH_CHARSET },
		{ "VIETNAMESE_CHARSET", VIETNAMESE_CHARSET },
		{ "JOHAB_CHARSET", JOHAB_CHARSET },
		{ "ARABIC_CHARSET", ARABIC_CHARSET },
		{ "HEBREW_CHARSET", HEBREW_CHARSET },
		{ "THAI_CHARSET", THAI_CHARSET },
		{ "OEM_CHARSET", OEM_CHARSET },
		{ 0, 0 }
		};

ButtonFont::ButtonFont()
{
	titleFont = 0;
	infoFont = 0;
	entryFont = 0;
	selentryFont = 0;
	folderFont = 0;
	selfolderFont = 0;
	arrowFont = 0;
	selarrowFont = 0;

	useTextArrow = false;
	arrowText = "";
	selarrowText = "";

	defaultfont.showText = true;

	defaultfont.leftBorder = 0;
	defaultfont.rightBorder = 0;
	defaultfont.topBorder = 0;
	defaultfont.bottomBorder = 0;

	defaultfont.shadow = false;
	defaultfont.shadowcolor = GetSysColor(COLOR_3DSHADOW);
	defaultfont.shadowX = 1;
	defaultfont.shadowY = 1;

	defaultfont.textoutline = false;
	defaultfont.textoutlinecolor = GetSysColor(COLOR_3DSHADOW);

	defaultfont.name = "Arial";
	defaultfont.charSet = DEFAULT_CHARSET;

	defaultfont.height = 15;
	defaultfont.color = RGB(0,0,0);

	defaultfont.bold = false;
	defaultfont.italic = false;
	defaultfont.underline = false;

	defaultfont.align = DT_LEFT;
	defaultfont.vertalign = DT_VCENTER;

	defaultfont.smoothing = true;
	defaultfont.clearType = true;
}

ButtonFont::~ButtonFont()
{
	if(titleFont)
		DeleteObject(titleFont);
	if(entryFont)
		DeleteObject(entryFont);
	if(selentryFont)
		DeleteObject(selentryFont);
	if(folderFont)
		DeleteObject(folderFont);
	if(selfolderFont)
		DeleteObject(selfolderFont);
	if(infoFont)
		DeleteObject(infoFont);
	if(arrowFont)
		DeleteObject(arrowFont);
	if(selarrowFont)
		DeleteObject(selarrowFont);
}

void ButtonFont::subconfigure(string type, FONTSETTINGS &workSettings, FONTSETTINGS &defaultSettings)
{
	workSettings.showText = GetxRCBool(type, "ShowText", defaultSettings.showText);
	
	workSettings.leftBorder = GetxRCInt(type, "FontLeftBorder", defaultSettings.leftBorder, 0);
	workSettings.rightBorder = GetxRCInt(type, "FontRightBorder", defaultSettings.rightBorder, 0);
	workSettings.topBorder = GetxRCInt(type, "FontTopBorder", defaultSettings.topBorder, 0);
	workSettings.bottomBorder = GetxRCInt(type, "FontBottomBorder", defaultSettings.bottomBorder, 0);

	workSettings.shadow = GetxRCBool(type, "FontShadow", defaultSettings.shadow);
	workSettings.shadowcolor = GetxRCColor(type, "FontShadowColor", defaultSettings.shadowcolor);
	workSettings.shadowX = GetxRCInt(type, "FontShadowX", defaultSettings.shadowX);
	workSettings.shadowY = GetxRCInt(type, "FontShadowY", defaultSettings.shadowY);

	workSettings.textoutline = GetxRCBool(type, "FontOutline", defaultSettings.textoutline);
	workSettings.textoutlinecolor = GetxRCColor(type, "FontOutlineColor", defaultSettings.textoutlinecolor);

	workSettings.name = GetxRCString(type, "Font", defaultSettings.name);
	workSettings.charSet = GetxRCNamedValue(type, "FontCharSet", CharSetConstants, defaultSettings.charSet );

	workSettings.height = GetxRCInt(type, "FontHeight", defaultSettings.height, 0);
	workSettings.color = GetxRCColor(type, "FontColor", defaultSettings.color);

	workSettings.bold = GetxRCBool(type, "FontBold", defaultSettings.bold);
	workSettings.italic = GetxRCBool(type, "FontItalic", defaultSettings.italic);
	workSettings.underline = GetxRCBool(type, "FontUnderline", defaultSettings.underline);

	workSettings.align = GetxRCNamedValue(type, "FontAlign", alignValues, defaultSettings.align);
	workSettings.vertalign = GetxRCNamedValue(type, "FontVertAlign", vertAlignValues, defaultSettings.vertalign);

	workSettings.smoothing = GetxRCBool(type, "FontSmoothing", defaultSettings.smoothing);
	workSettings.clearType = GetxRCBool(type, "FontClearType", defaultSettings.clearType);
}

void ButtonFont::configure(const string &name)
{
	subconfigure((name + "Title"), title, defaultfont);
	title.align = GetxRCNamedValue(name, "TitleFontAlign", alignValues, DT_CENTER);
	createHandle(title, titleFont);

	subconfigure((name + "Entry"), entry, defaultfont);
	createHandle(entry, entryFont);

	subconfigure((name + "ActiveEntry"), selentry, entry);
	createHandle(selentry, selentryFont);

	subconfigure((name + "Folder"), folder, entry);
	createHandle(folder, folderFont);

	subconfigure((name + "ActiveFolder"), selfolder, (GetxRCBool(name, "ActiveFolderUseFolderDefault") ? folder : selentry));
	createHandle(selfolder, selfolderFont);
	
	subconfigure((name + "Info"), info, defaultfont);
	info.align = GetxRCNamedValue(name, "InfoFontAlign", alignValues, DT_CENTER);
	createHandle(info, infoFont);

	useTextArrow = GetxRCBool(name, "UseTextArrow");
	if ( useTextArrow )
	{
		subconfigure((name + "Arrow"), arrow, defaultfont);
		createHandle(arrow, arrowFont);

		subconfigure((name + "ActiveArrow"), selarrow, arrow);
		createHandle(selarrow, selarrowFont);

		arrowText = GetxRCString(name, "ArrowText", "");
		selarrowText = GetxRCString(name, "ActiveArrowText", arrowText);
	}
}

POINT ButtonFont::measure(HDC hDC, string TextLine, int type)
{
	POINT maxValues = {0,0};
	POINT maxBorderValues = {0,0};
	LPSIZE lpsize = new SIZE;

	if (type == 0)
	{
		for (int i = 0; i < 4; i++)
		{
			if (i == 0)
			{
				if (!entry.showText)
					continue;
				entryFont = (HFONT) SelectObject(hDC, entryFont);
				if (maxBorderValues.x < entry.leftBorder+entry.rightBorder)
					maxBorderValues.x = entry.leftBorder+entry.rightBorder;
				if (maxBorderValues.y < entry.topBorder+entry.bottomBorder)
					maxBorderValues.y = entry.topBorder+entry.bottomBorder;
			}
			else if (i == 1)
			{
				if (!selentry.showText)
					continue;
				selentryFont = (HFONT) SelectObject(hDC, selentryFont);
				if (maxBorderValues.x < selentry.leftBorder+selentry.rightBorder)
					maxBorderValues.x = selentry.leftBorder+selentry.rightBorder;
				if (maxBorderValues.y < selentry.topBorder+selentry.bottomBorder)
					maxBorderValues.y = selentry.topBorder+selentry.bottomBorder;
			}
			else if (i == 2)
			{
				if (!folder.showText)
					continue;
				folderFont = (HFONT) SelectObject(hDC, folderFont);
				if (maxBorderValues.x < folder.leftBorder+folder.rightBorder)
					maxBorderValues.x = folder.leftBorder+folder.rightBorder;
				if (maxBorderValues.y < folder.topBorder+folder.bottomBorder)
					maxBorderValues.y = folder.topBorder+folder.bottomBorder;
			}
			else if (i == 3)
			{
				if (!selfolder.showText)
					continue;
				selfolderFont = (HFONT) SelectObject(hDC, selfolderFont);
				if (maxBorderValues.x < selfolder.leftBorder+selfolder.rightBorder)
					maxBorderValues.x = selfolder.leftBorder+selfolder.rightBorder;
				if (maxBorderValues.y < selfolder.topBorder+selfolder.bottomBorder)
					maxBorderValues.y = selfolder.topBorder+selfolder.bottomBorder;
			}

			lpsize->cx=10;
			lpsize->cy=10;

			GetTextExtentPoint32(hDC, TextLine.c_str(), strlen(TextLine.c_str()), lpsize);

			if (i == 0)
				entryFont = (HFONT) SelectObject(hDC, entryFont);
			else if (i == 1)
				selentryFont = (HFONT) SelectObject(hDC, selentryFont);
			else if (i == 2)
				folderFont = (HFONT) SelectObject(hDC, folderFont);
			else if (i == 3)
				selfolderFont = (HFONT) SelectObject(hDC, selfolderFont);
				
			if (maxValues.x < lpsize->cx)
				maxValues.x=lpsize->cx;
			if (maxValues.y < lpsize->cy)
				maxValues.y=lpsize->cy;
		}
	}
	else if (type == 1)
	{
		titleFont = (HFONT) SelectObject(hDC, titleFont);
		
		maxBorderValues.x = title.leftBorder+title.rightBorder;
		maxBorderValues.y = title.topBorder+title.bottomBorder;

		lpsize->cx=10;
		lpsize->cy=10;

		GetTextExtentPoint32(hDC, TextLine.c_str(), strlen(TextLine.c_str()), lpsize);

		titleFont = (HFONT) SelectObject(hDC, titleFont);
	
		maxValues.x=lpsize->cx;
		maxValues.y=lpsize->cy;
	}
	else if (type == 2)
	{
		infoFont = (HFONT) SelectObject(hDC, infoFont);
	
		maxBorderValues.x = info.leftBorder+info.rightBorder;
		maxBorderValues.y = info.topBorder+info.bottomBorder;

		lpsize->cx=10;
		lpsize->cy=10;

		GetTextExtentPoint32(hDC, TextLine.c_str(), strlen(TextLine.c_str()), lpsize);

		infoFont = (HFONT) SelectObject(hDC, infoFont);
	
		maxValues.x=lpsize->cx;
		maxValues.y=lpsize->cy;
	}

    delete lpsize;

	maxValues.x = maxValues.x+maxBorderValues.x;
	maxValues.y = maxValues.y+maxBorderValues.y;

	return maxValues;
}

void ButtonFont::apply(HDC hDC, int x, int y, int width, int height, const string &caption, int buttonType)
{
	RECT r;

	string drawText = caption;

	int textColor;
	int vertAlign;
	int align;

	int bkMode = SetBkMode(hDC, TRANSPARENT);

	if ( buttonType == 0 )
	{
		r.left = x + title.leftBorder;
		r.top = y + title.topBorder;
		r.right = x + width - title.rightBorder;
		r.bottom = y + height - title.bottomBorder;

		titleFont = (HFONT) SelectObject(hDC, titleFont);
		textColor = SetTextColor(hDC, title.color);
		vertAlign = title.vertalign;
		align = title.align;

		if (title.shadow && !title.textoutline)
		{
			OffsetRect(&r, (title.shadowX + 1) / 2, (title.shadowY + 1) / 2);
			SetTextColor(hDC, title.shadowcolor);
			DrawTextEx(hDC, (char *) drawText.c_str(), drawText.length(), &r, align | DT_SINGLELINE | vertAlign | DT_END_ELLIPSIS, 0);
			OffsetRect(&r, -title.shadowX, -title.shadowY);
			SetTextColor(hDC, title.color);
		}
		if (title.textoutline)
		{
			SetTextColor(hDC, title.textoutlinecolor);
			drawOutline(hDC, r, align, vertAlign, drawText);
			SetTextColor(hDC, title.color);
		}
	}
	else if ( buttonType == 1 )
	{
		if (!entry.showText)
			return;

		r.left = x + entry.leftBorder;
		r.top = y + entry.topBorder;
		r.right = x + width - entry.rightBorder;
		r.bottom = y + height - entry.bottomBorder;

		entryFont = (HFONT) SelectObject(hDC, entryFont);
		textColor = SetTextColor(hDC, entry.color);
		vertAlign = entry.vertalign;
		align = entry.align;

		if (entry.shadow && !entry.textoutline)
		{
			OffsetRect(&r, (entry.shadowX + 1) / 2, (entry.shadowY + 1) / 2);
			SetTextColor(hDC, entry.shadowcolor);
			DrawTextEx(hDC, (char *) drawText.c_str(), drawText.length(), &r, align | DT_SINGLELINE | vertAlign | DT_END_ELLIPSIS, 0);
			OffsetRect(&r, -entry.shadowX, -entry.shadowY);
			SetTextColor(hDC, entry.color);
		}
		if (entry.textoutline)
		{
			SetTextColor(hDC, entry.textoutlinecolor);
			drawOutline(hDC, r, align, vertAlign, drawText);
			SetTextColor(hDC, entry.color);
		}
	}
	else if ( buttonType == 2 )
	{
		if (!selentry.showText)
			return;

		r.left = x + selentry.leftBorder;
		r.top = y + selentry.topBorder;
		r.right = x + width - selentry.rightBorder;
		r.bottom = y + height - selentry.bottomBorder;

		selentryFont = (HFONT) SelectObject(hDC, selentryFont);
		textColor = SetTextColor(hDC, selentry.color);
		vertAlign = selentry.vertalign;
		align = selentry.align;

		if (selentry.shadow && !selentry.textoutline)
		{
			OffsetRect(&r, (selentry.shadowX + 1) / 2, (selentry.shadowY + 1) / 2);
			SetTextColor(hDC, selentry.shadowcolor);
			DrawTextEx(hDC, (char *) drawText.c_str(), drawText.length(), &r, align | DT_SINGLELINE | vertAlign | DT_END_ELLIPSIS, 0);
			OffsetRect(&r, -selentry.shadowX, -selentry.shadowY);
			SetTextColor(hDC, selentry.color);
		}
		if (selentry.textoutline)
		{
			SetTextColor(hDC, selentry.textoutlinecolor);
			drawOutline(hDC, r, align, vertAlign, drawText);
			SetTextColor(hDC, selentry.color);
		}
	}
	else if ( buttonType == 3 )
	{
		if (!folder.showText)
			return;

		r.left = x + folder.leftBorder;
		r.top = y + folder.topBorder;
		r.right = x + width - folder.rightBorder;
		r.bottom = y + height - folder.bottomBorder;

		folderFont = (HFONT) SelectObject(hDC, folderFont);
		textColor = SetTextColor(hDC, folder.color);
		vertAlign = folder.vertalign;
		align = folder.align;

		if (folder.shadow && !folder.textoutline)
		{
			OffsetRect(&r, (folder.shadowX + 1) / 2, (folder.shadowY + 1) / 2);
			SetTextColor(hDC, folder.shadowcolor);
			DrawTextEx(hDC, (char *) drawText.c_str(), drawText.length(), &r, align | DT_SINGLELINE | vertAlign | DT_END_ELLIPSIS, 0);
			OffsetRect(&r, -folder.shadowX, -folder.shadowY);
			SetTextColor(hDC, folder.color);
		}
		if (folder.textoutline)
		{
			SetTextColor(hDC, folder.textoutlinecolor);
			drawOutline(hDC, r, align, vertAlign, drawText);
			SetTextColor(hDC, folder.color);
		}
	}
	else if ( buttonType == 4 )
	{
		if (!selfolder.showText)
			return;

		r.left = x + selfolder.leftBorder;
		r.top = y + selfolder.topBorder;
		r.right = x + width - selfolder.rightBorder;
		r.bottom = y + height - selfolder.bottomBorder;

		selfolderFont = (HFONT) SelectObject(hDC, selfolderFont);
		textColor = SetTextColor(hDC, selfolder.color);
		vertAlign = selfolder.vertalign;
		align = selfolder.align;

		if (selfolder.shadow && !selfolder.textoutline)
		{
			OffsetRect(&r, (selfolder.shadowX + 1) / 2, (selfolder.shadowY + 1) / 2);
			SetTextColor(hDC, selfolder.shadowcolor);
			DrawTextEx(hDC, (char *) drawText.c_str(), drawText.length(), &r, align | DT_SINGLELINE | vertAlign | DT_END_ELLIPSIS, 0);
			OffsetRect(&r, -selfolder.shadowX, -selfolder.shadowY);
			SetTextColor(hDC, selfolder.color);
		}
		if (selfolder.textoutline)
		{
			SetTextColor(hDC, selfolder.textoutlinecolor);
			drawOutline(hDC, r, align, vertAlign, drawText);
			SetTextColor(hDC, selfolder.color);
		}
	}
	else if ( buttonType == 5 )
	{
		r.left = x + info.leftBorder;
		r.top = y + info.topBorder;
		r.right = x + width - info.rightBorder;
		r.bottom = y + height - info.bottomBorder;

		infoFont = (HFONT) SelectObject(hDC, infoFont);
		textColor = SetTextColor(hDC, info.color);
		vertAlign = info.vertalign;
		align = info.align;

		if (info.shadow && !info.textoutline)
		{
			OffsetRect(&r, (info.shadowX + 1) / 2, (info.shadowY + 1) / 2);
			SetTextColor(hDC, info.shadowcolor);
			DrawTextEx(hDC, (char *) drawText.c_str(), drawText.length(), &r, align | DT_SINGLELINE | vertAlign | DT_END_ELLIPSIS, 0);
			OffsetRect(&r, -info.shadowX, -info.shadowY);
			SetTextColor(hDC, info.color);
		}
		if (info.textoutline)
		{
			SetTextColor(hDC, info.textoutlinecolor);
			drawOutline(hDC, r, align, vertAlign, drawText);
			SetTextColor(hDC, info.color);
		}
	}
	else if ( buttonType == 31 )
	{
		drawText = arrowText;

		r.left = x + arrow.leftBorder;
		r.top = y + arrow.topBorder;
		r.right = x + width - arrow.rightBorder;
		r.bottom = y + height - arrow.bottomBorder;

		arrowFont = (HFONT) SelectObject(hDC, arrowFont);
		textColor = SetTextColor(hDC, arrow.color);
		vertAlign = arrow.vertalign;
		align = arrow.align;

		if (arrow.shadow && !arrow.textoutline)
		{
			OffsetRect(&r, (arrow.shadowX + 1) / 2, (arrow.shadowY + 1) / 2);
			SetTextColor(hDC, arrow.shadowcolor);
			DrawTextEx(hDC, (char *) drawText.c_str(), drawText.length(), &r, align | DT_SINGLELINE | vertAlign | DT_END_ELLIPSIS, 0);
			OffsetRect(&r, -arrow.shadowX, -arrow.shadowY);
			SetTextColor(hDC, arrow.color);
		}
		if (arrow.textoutline)
		{
			SetTextColor(hDC, arrow.textoutlinecolor);
			drawOutline(hDC, r, align, vertAlign, drawText);
			SetTextColor(hDC, arrow.color);
		}
	}
	else if ( buttonType == 41 )
	{
		drawText = selarrowText;

		r.left = x + selarrow.leftBorder;
		r.top = y + selarrow.topBorder;
		r.right = x + width - selarrow.rightBorder;
		r.bottom = y + height - selarrow.bottomBorder;

		selarrowFont = (HFONT) SelectObject(hDC, selarrowFont);
		textColor = SetTextColor(hDC, selarrow.color);
		vertAlign = selarrow.vertalign;
		align = selarrow.align;

		if (selarrow.shadow && !selarrow.textoutline)
		{
			OffsetRect(&r, (selarrow.shadowX + 1) / 2, (selarrow.shadowY + 1) / 2);
			SetTextColor(hDC, selarrow.shadowcolor);
			DrawTextEx(hDC, (char *) drawText.c_str(), drawText.length(), &r, align | DT_SINGLELINE | vertAlign | DT_END_ELLIPSIS, 0);
			OffsetRect(&r, -selarrow.shadowX, -selarrow.shadowY);
			SetTextColor(hDC, selarrow.color);
		}
		if (selarrow.textoutline)
		{
			SetTextColor(hDC, selarrow.textoutlinecolor);
			drawOutline(hDC, r, align, vertAlign, drawText);
			SetTextColor(hDC, selarrow.color);
		}
	}

	DrawTextEx(hDC, (char *) drawText.c_str(), drawText.length(), &r, align | DT_SINGLELINE | vertAlign | DT_END_ELLIPSIS, 0);
	
	if (buttonType == 0)
		titleFont = (HFONT) SelectObject(hDC, titleFont);
	else if (buttonType == 1)
		entryFont = (HFONT) SelectObject(hDC, entryFont);
	else if (buttonType == 2)
		selentryFont = (HFONT) SelectObject(hDC, selentryFont);
	else if (buttonType == 3)
		folderFont = (HFONT) SelectObject(hDC, folderFont);
	else if (buttonType == 4)
		selfolderFont = (HFONT) SelectObject(hDC, selfolderFont);
	else if (buttonType == 5)
		infoFont = (HFONT) SelectObject(hDC, infoFont);
	else if (buttonType == 31)
		arrowFont = (HFONT) SelectObject(hDC, arrowFont);
	else if (buttonType == 41)
		selarrowFont = (HFONT) SelectObject(hDC, selarrowFont);

	SetBkMode(hDC, bkMode);
	SetTextColor(hDC, textColor);
}

void ButtonFont::createHandle(FONTSETTINGS processFont, HFONT &processhfont)
{
	LOGFONT lf;
	memset(&lf, 0, sizeof(LOGFONT));

	StringCchCopy(lf.lfFaceName, LF_FACESIZE, processFont.name.c_str());
	lf.lfHeight = processFont.height;
	lf.lfItalic = processFont.italic ? TRUE : FALSE;
	lf.lfUnderline = processFont.underline ? TRUE : FALSE;
	lf.lfWeight = processFont.bold ? FW_BOLD : FW_NORMAL;
    lf.lfCharSet = processFont.charSet;
	if ( IsOS( OS_XP ) && processFont.smoothing && processFont.clearType )
		lf.lfQuality = CLEARTYPE_QUALITY;
	else if ( !processFont.smoothing )
		lf.lfQuality = NONANTIALIASED_QUALITY;

	if(processhfont != 0)
		DeleteObject(processhfont);

	processhfont = CreateFontIndirect(&lf);
}

void ButtonFont::drawOutline(HDC destHDC, RECT tmpr, int align, int vertAlign, const string &caption)
{
	OffsetRect(&tmpr, 1, 1);
	DrawTextEx(destHDC, (char *) caption.c_str(), caption.length(), &tmpr, align | DT_SINGLELINE | vertAlign | DT_END_ELLIPSIS, 0);

	OffsetRect(&tmpr, -1, 0);
	DrawTextEx(destHDC, (char *) caption.c_str(), caption.length(), &tmpr, align | DT_SINGLELINE | vertAlign | DT_END_ELLIPSIS, 0);

	OffsetRect(&tmpr, -1, 0);
	DrawTextEx(destHDC, (char *) caption.c_str(), caption.length(), &tmpr, align | DT_SINGLELINE | vertAlign | DT_END_ELLIPSIS, 0);

	OffsetRect(&tmpr, 0, -1);
	DrawTextEx(destHDC, (char *) caption.c_str(), caption.length(), &tmpr, align | DT_SINGLELINE | vertAlign | DT_END_ELLIPSIS, 0);

	OffsetRect(&tmpr, 0, -1);
	DrawTextEx(destHDC, (char *) caption.c_str(), caption.length(), &tmpr, align | DT_SINGLELINE | vertAlign | DT_END_ELLIPSIS, 0);

	OffsetRect(&tmpr, 1, 0);
	DrawTextEx(destHDC, (char *) caption.c_str(), caption.length(), &tmpr, align | DT_SINGLELINE | vertAlign | DT_END_ELLIPSIS, 0);

	OffsetRect(&tmpr, 1, 0);
	DrawTextEx(destHDC, (char *) caption.c_str(), caption.length(), &tmpr, align | DT_SINGLELINE | vertAlign | DT_END_ELLIPSIS, 0);

	OffsetRect(&tmpr, 0, 1);
	DrawTextEx(destHDC, (char *) caption.c_str(), caption.length(), &tmpr, align | DT_SINGLELINE | vertAlign | DT_END_ELLIPSIS, 0);
}
