/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2004 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#if !defined(AFX_POPUP_H__A3BA75CD_5C2F_484D_B3EC_CE64CFFF9734__INCLUDED_)
#define AFX_POPUP_H__A3BA75CD_5C2F_484D_B3EC_CE64CFFF9734__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <shlobj.h>

class PopupTexture;
class ButtonTexture;
class ButtonIcon;
class ButtonFont;

//Alpha Transparency
#ifndef WS_EX_LAYERED
	#define WS_EX_LAYERED           0x00080000L
	#define LWA_COLORKEY            0x00000001L
	#define ULW_COLORKEY            0x00000001L
	#define LWA_ALPHA               0x00000002L
	#define ULW_ALPHA               0x00000002L
	#define ULW_OPAQUE              0x00000003L
#endif // WS_EX_LAYERED

#ifndef AC_SRC_ALPHA
	#define AC_SRC_ALPHA            0x00000001L
#endif

typedef vector<HWND> hWndTaskList;
typedef vector<HWND>::iterator hWndTaskListIterator;

class ENTRY
{
public:

	ENTRY()
	{	
		hWnd = NULL;
		clsid = 0;
		pidl = NULL;
		iconPath = "";
		pipedList.clear();
		x = 0;
		y = 0;
		width = 0;
		height = 0;
		caption = "";
		keyCode = -1;
		keyCode1 = -1;
		action = "";
		info = "";
		entryType = 0;
		hiddenPart = 0;
		multiPart = 0;
	};

	virtual ~ENTRY()
	{
		pipedList.clear();
	};

	//Tasks and TasksParentEntry
	HWND hWnd;
	//ShellFolder ParentEntry
	int clsid;
	//ShellFolder Entrys
	LPITEMIDLIST pidl;
	//Normal Entrys
	string iconPath;
	//Multiple Folders with the same name
	StringList pipedList;
	int x;
	int y;
	int width;
	int height;
	//Displayed Text
	string caption;
	//Key for Keyboard Navigation
	int keyCode;
	int keyCode1;
	//Action For Files, Folders and Bangs
	string action;
	//Base Text, which is send to xLabel TextProcessing
	string info;
	//Type: -2 DynamicFolder
	//      -1 Folder
	//       0 Normal Entry
	//       1 Separator
	//       2 Info Entry
	int entryType;
	int hiddenPart;
	int multiPart;
};

typedef vector<ENTRY *>EntryList;
typedef vector<ENTRY *>::iterator EntryListIterator;

LPITEMIDLIST CreateIDList(int);
LPITEMIDLIST DuplicateIDList(LPCITEMIDLIST);
LPITEMIDLIST JoinIDLists(LPCITEMIDLIST, LPCITEMIDLIST);
int GetIDListSize(LPCITEMIDLIST);
LPCITEMIDLIST NextID(LPCITEMIDLIST);
void FreeIDList(LPITEMIDLIST);

class Popup  
{

public:

	Popup(const string &name, string subname = "", PopupTexture *parentpopuptexture = NULL, ButtonTexture *parentbuttontexture = NULL,	ButtonIcon *parentbuttonicon = NULL, ButtonFont *parentbuttonfont = NULL);
	virtual ~Popup();

	void reconfigure();
	void load(HINSTANCE hInstance);

	void ParseStepRC(bool externCall = false);
	StringList popupcontentList;

	void reposition(int type, int newx, int newy, int newwidth, int newheight, int externCall = 0);
	
	const HWND &gethWnd() const { return hWnd; };
	const string &getName() const { return name; };
	const int &getX() const { return actualx; };
	const int &getY() const { return actualy; };
	const int &getWidth() const { return actualwidth; };
	const int &getHeight() const { return actualheight; };
	const bool &isVisible() const { return visible; };

	void setAlwaysOnTop(bool newalwaysOnTop);

	void show(bool setcursor = false);
	void hide(bool noFade = false);

	void ToggleShade();

	string BangToCall;
	string captionname;

	void ScanSubMenu(string scanpath, int dynamic);
	void ScanShellFolder(int clsid);
	void ScanTasks();

	bool PinnedMode;
	bool pinnedPopupNotOnTop;

	bool openToLeft;

private:

	bool moveButtonDown;

	bool transparencyNeededAlways;

	RECT customWorkArea;
	bool restrictToWorkArea;

	bool hasHookedWindows;
	bool hasInfoEntry;

	bool ShadeAllowed;
	bool HideAllowed;
	bool MoveAllowed;

	bool AutoPin;

	bool alwaysOnTop;

	bool dropShadow;

	HWND ToolTips;

	bool autoMenuBreak;

	int quicklaunchMode;
	int quicklaunchItemSize;
	int quicklaunchWrapCount;
	int quicklaunchOpenTo;

	string openSound;
	string closeSound;
	string subSound;
	string execSound;

	bool execSoundPlayed;

	string dropCommand;
	string openCommand;
	string closeCommand;
	string shadeCommand;
	string unshadeCommand;

	int hideDelay;
	int trackDelay;
	int openDelay;
	int tasksUpdate;

	bool manualMove;

	int currentPart;
	int maxPart;
	int maxMenuBreak;

	EntryList entrylist;

	Popup* pParentPopup;
	ENTRY* pParentEntry;

	bool dynamicPopup;
	string rootPath;

	PopupList subpopupList;

	ENTRY* startentry;
	//Temp Pointers (Change often)
	ENTRY* lastactiveentry;
	ENTRY* hoverentry;
	
	int lastkeyentry;
	
	int entryHeight;
	int folderHeight;
	int infoHeight;
	int separatorHeight;

	int xSpacing;
	int ySpacing;

	bool autoSeparator;
	bool showExtensions;
	StringList noshowExtensions;

	int overlapX;
	int overlapY;

	bool folderOpenOnClick;

	void CreateSubMenu( string newsubname, string newrootpath, ENTRY* pNewParentEntry, bool jumpTo = false);
	void CreateShellSubMenu( string newsubname, ENTRY* pNewParentEntry, int clsid, bool jumpTo = false);
	void CreateTaskSubMenu( string newsubname, ENTRY* pNewParentEntry, bool jumpTo = false);
	
	void Invoke(LPITEMIDLIST m_pidlFull);
	void RInvoke(LPITEMIDLIST m_pidlFull);

	int shaded;

	int titleLClickAction;
	int titleMClickAction;
	int titleRClickAction;

	bool isTaskPopup;
	hWndTaskList tasklist;
	
	void ActionAtPoint( int xP, int yP, bool execute = false, bool context = false);
	void HoverActions( int xP, int yP );

	bool SelectNextEntry(bool upwards, bool noCursorMove = false);

	bool PopupActive();

	string name;
	string subname;

	HWND hWnd;
	HINSTANCE hInstance;

	HDC defaultDC;
	HBITMAP defaultBitmap;

	string DefaultIcon;
	string DefaultFolderIcon;

	int actualx;
	int actualy;
	int actualwidth;
	int actualheight;

	int leftBorder;
	int topBorder;
	int rightBorder;
	int bottomBorder;

	int maxWidth;
	int maxHeight;
	int minWidth;
	int minHeight;

	int orgminHeight;

	int maxEntryWidth;
	int minEntryWidth;

	bool visible;

	bool mouseInside;
	bool mousePressed;
	bool mouseMoved;
	POINT savedpt;

	bool alphaFade;
	int UpdateAlpha;
	int alphaFadeStep;
	int alphaFadeDelay;

	int alphaTransparency;
	bool alphaMap;

	int showCaption;

	void repaint(bool invalidateCache = false, bool regionUpdate = false);
	void updateRegion();

	PopupTexture *popuptexture;
	ButtonTexture *buttontexture;
	ButtonIcon *buttonicon;
	ButtonFont *buttonfont;

	POINT oldmovetracker;

	int entryCount;

	bool inChordingTimeOut;
	
private:

	void UpdateWindowTransparent(HWND hWnd, int factor);

	int isCloseTo(POINT &pos, const POINT size, int x, int y, const int cx, const int cy, const int span);

	void Sort(vector<string> &inputList);
	void SortTasks(hWndTaskList &inputList);
	void SortShellFolder(vector<LPITEMIDLIST> &inputList);

	static bool Compare(string m1, string m2)
	{
		CString helper1 = m1.c_str();
		CString helper2 = m2.c_str();
		helper1 = helper1.Right(helper1.GetLength()-helper1.ReverseFind('\\'));
		helper2 = helper2.Right(helper2.GetLength()-helper2.ReverseFind('\\'));
		
		if (lstrcmpi(helper1, helper2) < 0)
			return true;
		else
			return false;
	};
	static bool CompareTasks(HWND m1, HWND m2)
	{
		char windowCaption1[MAX_PATH_LENGTH] = {0};
		char windowCaption2[MAX_PATH_LENGTH] = {0};
	
		GetWindowText( m1, windowCaption1, MAX_PATH_LENGTH );
		GetWindowText( m2, windowCaption2, MAX_PATH_LENGTH );

		if (lstrcmpi(windowCaption1, windowCaption2) < 0)
		{
			return true;
		}
		else
		{
			return false;
		}
	};
	static bool CompareShellFolders(LPITEMIDLIST p1, LPITEMIDLIST p2)
	{
		SHFILEINFO shfi1;
		SHGetFileInfo((LPCSTR) p1, 0, &shfi1, sizeof(SHFILEINFO),
										  SHGFI_DISPLAYNAME | SHGFI_PIDL);

		SHFILEINFO shfi2;
		SHGetFileInfo((LPCSTR) p2, 0, &shfi2, sizeof(SHFILEINFO),
										  SHGFI_DISPLAYNAME | SHGFI_PIDL);

		if (lstrcmpi(shfi1.szDisplayName, shfi2.szDisplayName) < 0)
			return true;
		else
			return false;
	};
	
	string InfoUpdate(string infocfg);

	void updateHint(HWND hWnd, LPSTR caption, bool updatesize, bool remove = false);

	void ScanKeyCodes(ENTRY* processEntry, string processString);


protected:

	void ontemplateButtonDown(const int &x, const int &y, int actionType);
	void ontemplateButtonUp(const int &x, const int &y, int actionType, int mousebutton);

	virtual void onLButtonDblClk(int x, int y);
	virtual void onWheelDown(int x, int y);
	virtual void onWheelUp(int x, int y);
	virtual void onKey(int nKey);
	virtual void onMouseMove(int x, int y);
	virtual void onPaint(HDC hDC);
	virtual void onWindowPosChanging(WINDOWPOS *windowPos);
	virtual void onMove(int x, int y);;
	virtual void onSize(int width, int height);
	virtual void onTimer(int timerID);

private:
	BOOL onWindowMessage( UINT message, WPARAM wParam, LPARAM lParam, LRESULT &lResult );

private:   // statics
	static BOOL WINAPI EnumWindowsProc( HWND hWnd, LPARAM lParam );

public:
	static LRESULT WINAPI windowProcedure(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
};

#endif // !defined(AFX_POPUP_H__A3BA75CD_5C2F_484D_B3EC_CE64CFFF9734__INCLUDED_)
